/**
 *** prog84.c
 ***
 *** these routines use the parallel or serial port
 *** to program a PIC 16C84 in serial mode
 ***/

#include <string.h>
#include <stdlib.h>
#include "prog84.h"

int
    p_power   = -1,
    p_mclr    = -1,
    p_clock   = -1,
    p_data    = -1,
    p_clock_f = -1,
    p_data_f  = -1;
int
    p_loopcnt = -1;

struct lp_io lpb;

/* if you edit this array be sure to change the array lengths in the
   various calls that reference it */
static 
struct lp_name names[] = {
    { "base",    lpc_number, &lpb.base },
    { "loop",    lpc_number, &p_loopcnt},

    { "power",   lpc_bits,   &p_power    },
    { "mclr",    lpc_bits,   &p_mclr     },
    { "clock",   lpc_bits,   &p_clock    },
    { "data",    lpc_bits,   &p_data     },
    { "clock_f", lpc_bits,   &p_clock_f  },
    { "data_f",  lpc_bits,   &p_data_f   },

    /* alternate names */
    { "rb6",    lpc_abits,   &p_clock    },
    { "rb7",    lpc_abits,   &p_data     },
    { "rb6_f",  lpc_abits,   &p_clock_f  },
    { "rb7_f",  lpc_abits,   &p_data_f   },
    { ""     ,  lpc_eof,     NULL        }

};

char *cfg_file = "lp_cfg";
char *progpath="" ;

int chiptype=PIC16x84;

chip_info *chip_p, 
	chips[]=   
	      { {"16x84",PIC,0x400,64},
		{"24c16",EEPROM,0,0x800}, 
		{"24c08",EEPROM,0,0x400},
		{"24c65",EEPROM,0,0x400}};

int p84_verbose = 0;
int i2c_debug=0;
unsigned i2cdelay=0;   /* if CPU's get to fast in the future for the 24c16 */

static int file_exist(char *s) {
  FILE *f;
  if (s==NULL) return 0;
  if ( (f=fopen(s,"r"))==NULL) return 0;
  fclose(f);
  return 1;
}

/* set program path from argv */
void set_progpath(char *p) {
  char *pr;
  pr=strrchr(p,'/');
  if (pr!=NULL) {  
	int  l;
	l=strlen(p)-strlen(pr);
	progpath=calloc(l+2,sizeof(char));
	strncpy(progpath,p,l+1);
	pr=calloc(l+2+strlen(cfg_file),sizeof(char));
	strcpy(pr,progpath);
	strcat(pr,cfg_file);
  } else pr="";
 /* find lp_cfg:  
    1st.: currentdir; 2nd.: $HOME/.lp_cfg; 3rd.:"progpath"/lp_cfg */
  if ( ! file_exist(cfg_file) ) {
#ifdef __MSDOS__
    if ( (strlen(pr)>0) && file_exist(pr)) cfg_file=pr;
#else
    char *home;
    char *cf;
    home = getenv("HOME");
    if (home != NULL) {
	cf=calloc(strlen(cfg_file)+strlen(home)+3,sizeof(char));
	strcpy(cf,home); strcat(cf,"/.");
	strcat(cf,cfg_file);
    } else cf=pr;
    if (file_exist(cf)) cfg_file=cf;
    else if (file_exist(pr)) cfg_file=pr;
#endif
  }
}

/***
 *** poweron the pic and enter programming mode
 ***/
void power_and_reset()
{

    /* clear power, if it still on from previos call */
    SET(mclr, 0);
    SET(power, 0);
    SET(clock, 1);
    SET(data, 0);
    FLUSH;
    usleep(200000UL); /* 0,2 s is hopefully enough to reset programmer */

    /* set power to pic-programmer */
    SET(mclr, 0);
    SET(power, 1);	/* the serial programmer needs this... */
    SET(clock, 0);
    SET(data, 0);
    FLUSH;
    if (p84_verbose) printf("Charging capacitor (2 sec.)...\n");
    sleep(2); /* wait quite a bit to charge capacitor */
    SET(mclr, 0);
    FLUSH;
    usleep(50000UL);
    SET(mclr, 1);
    FLUSH;
}

int test_chip() {
    int chiptp=PIC16x84;
    i2c_start();
    if (i2c_sendbyte((unsigned char) 0xa0)) {
	chiptp=EE24C65;
	i2c_start();
	if (i2c_sendbyte((unsigned char) 0xae)) chiptp=EE24C16;
	else if (i2c_sendbyte((unsigned char) 0xa6)) chiptp=EE24C08;
    };
    fprintf(stderr,"\n%s", chiptp==PIC16x84 ? "Assuming" : "Detected");
    fprintf(stderr," %s (%u) chip\n",chips[chiptp].name,chiptp);
    chip_p=&chips[chiptp];
    return chiptp;
}

int testHW()
{
    int i,j;     /* utility variables */
    int ret[4];  /* array of test results */
    int m;       /* merged test results */

    power_and_reset();

    if (p_power == -1 || p_mclr == -1 || p_clock == -1 ||
	    p_data == -1 || p_data_f == -1) {
	fprintf(stderr, 
	    "Error: I need control over power, MCLR, clock, and data,\n"
	    "       and I need to be able to read data back.\n");
	return 1;
    }


    if (p84_verbose) {
	printf("[testing loopback   ");
	fflush(stdout);
    }

    for (i=0; i<4; i++) {
	SET(clock, i&1);
	SET(data, i&2);
	FLUSH;
	usleep(100UL); /* here I can be slow, no problem */
	j = GET(data_f) ? 2 : 0;
	if (p_clock_f != -1)
	    j |= GET(clock_f)? 1 : 0;
	else
	    j |= i & 1;  /* fake it if we don't have it */
	ret[i] = j;
    }

    SET(clock, 0);
    SET(data, 0);
    FLUSH;

    if (p84_verbose)
	printf("  ]\n");

    /***
     *** analyze the test results. We put our four 2-bit responses
     *** into an int, one per nybble so we can use handy hex notation.
     ***/

    m = (((((ret[3] << 4) | ret[2]) << 4) | ret[1]) << 4) | ret[0];

    if(m == 0x3210)	/*** this is what we expect ***/
	return 0;

    fprintf(stderr, "Bad test results!\n  [data, clock]->[data_f, clock_f]:");
    for(i = 0; i < 4; i ++)
	fprintf(stderr, " %d%d->%d%d", (i>>1)&1, i&1,
		(ret[i]>>1)&1, (ret[i])&1);
    fprintf(stderr,"\n");

    if (p_clock_f == -1)
	fprintf(stderr, "(clock_f simulated since I don't actually have it)\n");

    if (m == 0x3120)
	fprintf(stderr, "clock and data reversed?\n"
	    " (could be either the clock & data lines, or the sense lines)\n"
	);

    switch(m & 0x1111) {
    case 0x0000:
	fprintf(stderr, "clock_f stuck low?\n");
	break;
    case 0x0101:
	fprintf(stderr, "clock_f inverted?\n");
	break;
    case 0x1111:
	fprintf(stderr, "clock_f stuck high?\n");
	break;
    }

    switch(m & 0x2222) {
    case 0x0000:
	fprintf(stderr,"data_f stuck low?\n");
	break;
    case 0x0022:
	fprintf(stderr,"data_f inverted?\n");
	break;
    case 0x2222:
	fprintf(stderr,"data_f stuck high?\n");
	break;
    }

    /***
     *** here I evidently have a failure, and should return 1. However,
     *** I keep on trying in the hope it is a temporary thing.
     ***/

    
    return 0;
}


/*** 
 *** Pipe some bits through the 16C84. Returns the values read from
 *** the data pin at the appropriate point in the cycle for reading
 *** data mem.  In a write, will return 'bits'.  (Unless, of course,
 *** something's wrong.)
 *** When reading set bits=0x7ffe
 ***/

#ifdef __MSDOS__
int loopd1=2,loopd2=200;
#endif

unsigned int pipeBits(unsigned int bits, int count)
{
    unsigned int ret = 0;
    int bit;

    for (bit=0; bit<count; bit++) {
	SET(clock, 1);
	SET(data, bits&1);
	FLUSH;
	/* Minumum 100 ns. delay is needed when programming PIC's
	   specifications require min. 100ns clock low and 100ns clock low.
	   Data must be valid > 80ns before clock transition 1->0
	   FLUSH ( lpb_flush() , lp_refresh() ) usually
	   gives adequate delay on linux/frebsd systems */
	SLEEP100ns(loopd1); /* 0.2 us */

	SET(clock, 0);
	FLUSH;   /* <- loads data in PIC , & reads input */
	ret |= (GET(data_f)? 1 : 0 ) << bit;
	if (p84_verbose > 1)
	    printf(". bit %2d: send %d get %d\n",
		bit, bits&1, GET(data_f)?1:0);

	/* minimum 100ns delay needed : */
	SLEEP100ns(loopd1); /* 0.2us */
	bits >>= 1;
    }
    SET(data, 1); /* the HI-Z state */
    FLUSH;
    /* required guard time of at least 1 usec.
       seems for fast cpu + dos/win at least 1ms is needed !? */
    SLEEP100ns(loopd2); /* 2.0 us */
    return ret;
}

void xPipeBits(unsigned int bits, int count)
{
    unsigned int ret;

    ret = pipeBits(bits, count);
    if (ret != bits) {
	fprintf(stderr,
	    "ERROR: sent %d bits: %d (0%o), got %d (0%o)\n"
	    "Resetting & powering down.\n",
	    count, bits, bits, ret, ret);
	progShutdown();
	exit(2);
    }
}

/* A generic command-plus-14-bits-to-write call */
void genericWrite(int cmd, unsigned int data)
     /* 6 bits of command, 14 bits of data */
{
    xPipeBits(cmd, 6);		  	/* Send the command */
    xPipeBits((data << 1) & 0x7FFE, 16);	/* send the data */
}

/* A generic command-plus-14-bits-to-read call */
unsigned int genericRead(int cmd)
{
    unsigned int ret;

    xPipeBits(cmd, 6);			/* send the command */
    ret = pipeBits(0x7FFE, 16);		/* read the result */
    return((ret & 0x7FFE) >> 1);	/* remove guard bits */
}


void progSetup()
{
    /* read the configuration file */
    if (lpb_cfg_read(cfg_file, 12, names)) {
	fprintf(stderr, "error reading cfg file %s\n", cfg_file);
	exit(1);
    }

    if (p84_verbose)
	lpb_dump_names(8, names);

    /* open /dev/port */
    open_io();


    /* check that the programmer's not totally dead */
    if (testHW()) {
	fprintf(stderr,"\nProgrammer is not responding, aborting.\n");
	exit(1);
    }
    chiptype=test_chip();
    i2cdelay=0;
}

void progShutdown()
{
    if (verbose) fprintf(stderr,"...closing\n");
    SET(mclr, 0);
    FLUSH;
    usleep(10000UL);	/* shutdown... */
    SET(power, 0);
    SET(clock, 0);
    SET(data, 0);
    FLUSH;
    close_io();
}

/* 990103 Frank Damgaard , for eeprom :
   Partially borrowed from i2c driver in "bttv" package,
   and ideas from serp-0.5.tgz
   Datasheet for 24c16 from www.microchip.com was a good help too :)
   Support for 24c08 and 24c65 is not tested!
*/


void i2c_out(WORD c, WORD d) {
  SET(clock,c); SET(data,d); FLUSH;
  /* I2C_DEBUG(fprintf(stderr,"%c",d?'1':'0')); */
  if (i2cdelay) usleep((unsigned long) i2cdelay);  
}

void i2c_start()
{ /* the 1->0 transition while clock=1 => start */
        I2C_SET(0,1);
        I2C_SET(1,1);
        I2C_SET(1,0);
        I2C_SET(0,0);
        I2C_DEBUG(fprintf(stderr,"i2c_start()\n")) ;
}

void i2c_stop()
{ /* the 0->1 transition while clock=1 => stop */
        I2C_SET(0,0);
        I2C_SET(1,0);
        I2C_SET(1,1);
        I2C_DEBUG(fprintf(stderr,"i2c_stop()\n"));
}

void i2c_data(WORD short dat)
{ /* data is clocked during clock=1 */
        I2C_SET(0,dat);
        I2C_SET(1,dat);
        I2C_SET(0,dat);
}


int i2c_ack(int ack)
{ /* 
  ack=1 if acknowledge is wanted
  return:  0  if no acknowledge
           1  if acknowledge 
*/
     int rack;
      /* if data=ack=1 is sent after command, 
         an 0 is received if ok */
	I2C_SET(0,ack);
	I2C_SET(1,ack);
	REFRESH;
        rack = ! I2C_GET();
        I2C_SET(0,ack);
        return rack; 
}

int i2c_sendbyte(unsigned char dat) {
        int i, ack;
    
        I2C_SET(0,0);
        for (i=7; i>=0; i--) i2c_data(dat&(1<<i));
        ack=i2c_ack(I2C_ACK);
        I2C_DEBUG(fprintf(stderr," %02x%c ",(int)dat,ack?'+':'-'));
        return ack;
}

void i2c_sendctrl(unsigned char com, WORD adr) {
  unsigned char ctrl=0;
  if (chiptype==EE24C16) ctrl|=adr>>7 ;
  ctrl&=0x0e;
  ctrl|=com;
  i2c_sendbyte(ctrl);
}



int i2c_waitack(unsigned int t) {
/* block write and byte write need up to 10 milleseconds wait time. */
/* t in microseconds */
  int i,ack;
  
  if (t==0) t=1000; else t=t/1000;
  for (i=0; i<t; i++) {
	usleep(10UL);
	i2c_start();
	ack=i2c_sendbyte(EE_CTRL_RD);
	if (!ack) return 0;
  }
  return 1;
}

unsigned char i2c_readbyte(int ack) {
        int i;
        unsigned char dat=0;
    
        I2C_SET(0,1);
        for (i=7; i>=0; i--) {
                I2C_SET(1,1); REFRESH; 
                if (I2C_GET())
                        dat |= (1<<i);
                I2C_SET(0,1);
	}
	i=i2c_ack(ack); 
        I2C_DEBUG(fprintf(stderr,"=%02x%c%c ",(int)dat,ack?'+':'-',i?'+':'-'));
        return dat;
}

void i2c_sendaddr(WORD adr) {
  i2c_sendctrl(EE_CTRL_WR,adr);
  if (chiptype==EE24C65)  i2c_sendbyte((adr>>8)&0x1f);
  i2c_sendbyte(adr&0xff);
  I2C_DEBUG(fprintf(stderr," sendaddr=%04x ",(int)adr));
}

void i2c_seqread(WORD adr, unsigned char *data, int len) {
  int i;
  i2c_start();
  i2c_sendaddr(adr);
  i2c_start();
  i2c_sendctrl(EE_CTRL_RD, adr);
  for (i=0; i<len-1; i++) {
	data[i]=i2c_readbyte(I2C_NOACK);
  }
  data[len-1]=i2c_readbyte(I2C_ACK);
  i2c_stop();
}

unsigned char i2c_read(WORD adr) {
  unsigned char data;
  i2c_start();
  i2c_sendaddr(adr);
  i2c_start();
  i2c_sendctrl(EE_CTRL_RD, adr);
  data=i2c_readbyte(I2C_ACK);
  I2C_DEBUG(fprintf(stderr," data=%02x ",(int)data));
  i2c_stop();
  return data;
}


void i2c_pagewrite(WORD adr, unsigned char *dat, WORD PageSize) {
/* Must do an i2c_waitack() after i2c_pagewrite()! */
/* *dat must hold not more than 16 words (unsigned char) !! */
/* 24c16/c08 only has 16 byte storage so pagewrite() onlsy writes 16 bytes*/

  int i;
  i2c_start();
  i2c_sendaddr(adr);
  for (i=0; i<PageSize; i++) i2c_sendbyte(dat[i]);
  i2c_stop();
}

void i2c_write(WORD adr, WORD data) {
/* Must do an i2c_waitack() after i2c_write()! */
  i2c_start();
  i2c_sendaddr(adr);
  i2c_sendbyte(data);
  i2c_stop();
}
